from datetime import datetime

import pytz


def current_datetime_after_given_iso_string(value: str) -> bool:
    # Grab local timezone
    uk_tz = pytz.timezone("Europe/London")
    # Grab current localised datetime
    now_datetime = datetime.now(uk_tz).replace(tzinfo=None)
    # Convert value datetime to correct format for comparison
    parsed = datetime.fromisoformat(value)
    return now_datetime > parsed


def current_datetime_before_given_iso_string(value: str) -> bool:
    # Grab local timezone
    uk_tz = pytz.timezone("Europe/London")
    # Grab current localised datetime
    now_datetime = datetime.now(uk_tz).replace(tzinfo=None)
    # Convert value datetime to correct format for comparison
    parsed = datetime.fromisoformat(value)
    return now_datetime < parsed



from collections import defaultdict

from flask import current_app
from fsd_utils import NotifyConstants
from fsd_utils.mapping.application.application_utils import format_checkbox
from fsd_utils.mapping.application.application_utils import format_date_month_year
from fsd_utils.mapping.application.application_utils import format_month_year
from fsd_utils.mapping.application.application_utils import format_radio_field
from fsd_utils.mapping.application.application_utils import NO
from fsd_utils.mapping.application.application_utils import YES
from fsd_utils.mapping.application.free_text import FreeText
from fsd_utils.mapping.application.languages import EN
from fsd_utils.mapping.application.languages import SUPPORTED_LANGUAGES
from fsd_utils.mapping.application.multi_input import MultiInput


def extract_questions_and_answers(forms, language=EN) -> dict:
    if not language and language not in SUPPORTED_LANGUAGES:
        language = EN

    """function takes the form data and returns
    dict of questions & answers.
    """
    questions_answers = defaultdict(dict)
    try:
        for count, form in enumerate(forms, start=1):
            form_name = form["name"]
            current_app.logger.info(f"Form {count}) {form_name}")
            if form_name in form[NotifyConstants.APPLICATION_NAME_FIELD]:
                for question in form[NotifyConstants.APPLICATION_QUESTIONS_FIELD]:
                    for field in question["fields"]:
                        answer = field.get("answer")

                        if field["type"] == "file":
                            # we check if the question type is "file"
                            # then we remove the aws
                            # key attached to the answer
                            if isinstance(answer, str):
                                questions_answers[form_name][
                                    field["title"]
                                ] = answer.split("/")[-1]
                            else:
                                questions_answers[form_name][field["title"]] = answer

                        elif isinstance(answer, bool) and field["type"] == "list":
                            questions_answers[form_name][field["title"]] = (
                                YES[language] if answer else NO[language]
                            )

                        elif isinstance(answer, list) and field["type"] == "multiInput":
                            questions_answers[form_name][
                                field["title"]
                            ] = MultiInput.map_multi_input_data(answer, language)

                        elif field["type"] == "freeText":
                            clean_html_answer = FreeText.remove_html_tags(answer)

                            questions_answers[form_name][
                                field["title"]
                            ] = clean_html_answer

                        elif isinstance(answer, list) and field["type"] == "list":
                            questions_answers[form_name][
                                field["title"]
                            ] = format_checkbox(answer)

                        elif isinstance(answer, str) and field["type"] == "list":
                            questions_answers[form_name][
                                field["title"]
                            ] = format_radio_field(answer)

                        elif field["type"] == "monthYear":
                            questions_answers[form_name][
                                field["title"]
                            ] = format_month_year(answer)

                        elif field["type"] == "date":
                            questions_answers[form_name][
                                field["title"]
                            ] = format_date_month_year(answer)

                        else:
                            # we dont want to display boolean question
                            # (Do you want to mark this section as complete?)
                            if (
                                field.get("type") == "boolean"
                                and field.get("title").strip().lower()
                                == "do you want to mark this section as complete?"
                            ):
                                continue
                            else:
                                questions_answers[form_name][field["title"]] = answer
    except Exception as e:
        current_app.logger.error(f"Error occurred while processing form data: {e}")
        current_app.logger.error(f"Could not map the data for form: {form_name}")
        raise Exception(f"Could not map the data for form: {form_name}")
    return questions_answers



from uuid import uuid4

from fsd_utils.services.aws_sqs_extended_client_exception import ExceptionMessages
from fsd_utils.services.aws_sqs_extended_client_exception import (
    SQSExtendedClientException,
)

S3_KEY_ATTRIBUTE_NAME = "S3Key"
MAX_ALLOWED_ATTRIBUTES = 10 - 1  # 10 for SQS and 1 reserved attribute
DEFAULT_MESSAGE_SIZE_THRESHOLD = 262144
RESERVED_ATTRIBUTE_NAME = "ExtendedPayloadSize"
MESSAGE_POINTER_CLASS = "software.amazon.payloadoffloading.PayloadS3Pointer"


def get_message_attributes_size(message_attributes: dict) -> int:
    """
    Responsible for calculating the size, in bytes, of the message attributes
    of a given message payload
    :message_attributes: A dictionary consisting of message attributes.
    Each message attribute consists of the name (key) along with a
    type and value of the message body. The following types are supported
    for message attributes: StringValue, BinaryValue and DataType.
    """
    total_message_attributes_size = 0
    for key, value in message_attributes.items():
        total_message_attributes_size += len(key.encode("utf-8"))

        datatype_value = value.get("DataType", 0)
        if datatype_value:
            total_message_attributes_size += len(key.encode("utf-8"))

        stringtype_value = value.get("StringValue", 0)
        if stringtype_value:
            total_message_attributes_size += len(key.encode("utf-8"))

        binary_type_value = value.get("BinaryValue", 0)
        if binary_type_value:
            total_message_attributes_size += len(key.encode("utf-8"))

    return total_message_attributes_size


def get_reserved_attribute_name_if_present(message_attributes: dict) -> str:
    """
    Responsible for checking the reserved message attribute, SQSLargePayloadSize
    or ExtendedPayloadSize in this specific case, exists in the
    message_attributes
    :message_attributes: A dictionary consisting of message attributes.
    Each message attribute consists of the name (key) along with a
    type and value of the message body. The following types are supported
    for message attributes: StringValue, BinaryValue and DataType.
    """
    reserved_attribute_name = ""
    if RESERVED_ATTRIBUTE_NAME in message_attributes:
        reserved_attribute_name = RESERVED_ATTRIBUTE_NAME
    return reserved_attribute_name


def check_message_attributes(message_attributes: dict) -> None:
    """
    Responsible for checking the constraints on the message attributes
    for a given message
    :message_attributes A dictionary consisting of message attributes.
    Each message attribute consists of the name (key) along with a
    type and value of the message body. The following types are supported
    for message attributes: StringValue, BinaryValue and DataType.
    """
    total_message_attributes_size = get_message_attributes_size(message_attributes)

    if total_message_attributes_size > DEFAULT_MESSAGE_SIZE_THRESHOLD:
        raise SQSExtendedClientException(
            ExceptionMessages.INVALID_MESSAGE_ATTRIBUTE_SIZE.format(
                total_message_attributes_size, DEFAULT_MESSAGE_SIZE_THRESHOLD
            )
        )

    message_attributes_num = len(message_attributes)
    if message_attributes_num > MAX_ALLOWED_ATTRIBUTES:
        raise SQSExtendedClientException(
            ExceptionMessages.INVALID_NUMBER_OF_MESSAGE_ATTRIBUTES.format(
                message_attributes_num, MAX_ALLOWED_ATTRIBUTES
            )
        )

    reserved_attribute_name = get_reserved_attribute_name_if_present(message_attributes)
    if reserved_attribute_name:
        raise SQSExtendedClientException(
            ExceptionMessages.INVALID_ATTRIBUTE_NAME_PRESENT.format(
                reserved_attribute_name
            )
        )

    return


def get_s3_key(message_attributes: dict, extra_attributes: dict) -> str:
    """
    Responsible for checking if the S3 Key exists in the
    message_attributes
    :message_attributes: A dictionary consisting of message attributes
    :extra_attributes: A dictionary consisting of message attributes
    Each message attribute consists of the name (key) along with a
    type and value of the message body. The following types are supported
    for message attributes: StringValue, BinaryValue and DataType.
    """

    if S3_KEY_ATTRIBUTE_NAME in message_attributes:
        return message_attributes[S3_KEY_ATTRIBUTE_NAME]["StringValue"]
    elif extra_attributes and S3_KEY_ATTRIBUTE_NAME in extra_attributes:
        return (
            extra_attributes[S3_KEY_ATTRIBUTE_NAME]["StringValue"] + "/" + str(uuid4())
        )
    return str(uuid4())


def validate_messages(messages):
    for msg in messages:
        if not msg["MessageId"] and not msg["ReceiptHandle"] and not msg["Body"]:
            raise SQSExtendedClientException(
                ExceptionMessages.INVALID_ARGUMENTS_FOR_DELETE_MESSAGE
            )



import unittest
from unittest.mock import MagicMock
from uuid import uuid4

import boto3
from fsd_utils.sqs_scheduler.context_aware_executor import ContextAwareExecutor
from fsd_utils.sqs_scheduler.task_executer_service import TaskExecutorService
from moto import mock_aws


class TestTaskExecutorService(unittest.TestCase):
    @mock_aws
    def test_message_in_mock_environment_processing(self):
        """
        This test ensure that when message is there and if no errors occurred while processing the message
        then successfully removed it from the queue
        """
        self._mock_aws_client()
        self._add_data_to_queue()

        self.task_executor.process_messages()

        self._check_is_data_available(0)

    def _mock_aws_client(self):
        """
        Mocking aws resources and this will act as real aws environment behaviour
        """
        bucket_name = "fsd_msg_s3_bucket"
        self.flask_app = MagicMock()
        self.executor = ContextAwareExecutor(
            max_workers=10, thread_name_prefix="NotifTask", flask_app=self.flask_app
        )
        s3_connection = boto3.client(
            "s3",
            region_name="us-east-1",
            aws_access_key_id="test_accesstoken",  # pragma: allowlist secret
            aws_secret_access_key="secret_key",  # pragma: allowlist secret
        )
        sqs_connection = boto3.client(
            "sqs",
            region_name="us-east-1",
            aws_access_key_id="test_accesstoken",  # pragma: allowlist secret
            aws_secret_access_key="secret_key",  # pragma: allowlist secret
        )
        s3_connection.create_bucket(Bucket=bucket_name)
        self.queue_response = sqs_connection.create_queue(
            QueueName="notif-queue.fifo", Attributes={"FifoQueue": "true"}
        )
        self.task_executor = AnyTaskExecutorService(
            flask_app=MagicMock(),
            executor=self.executor,
            s3_bucket=bucket_name,
            sqs_primary_url=self.queue_response["QueueUrl"],
            task_executor_max_thread=5,
            sqs_batch_size=10,
            visibility_time=1,
            sqs_wait_time=2,
            endpoint_url_override=None,
            aws_access_key_id="test_accesstoken",  # pragma: allowlist secret
            aws_secret_access_key="secret_key",  # pragma: allowlist secret
            region_name="us-east-1",
        )
        self.task_executor.sqs_extended_client.sqs_client = sqs_connection
        self.task_executor.sqs_extended_client.s3_client = s3_connection

    def _add_data_to_queue(self):
        """
        Adding test data into the queue
        """
        for x in range(1):
            message_id = self.task_executor.sqs_extended_client.submit_single_message(
                queue_url=self.queue_response["QueueUrl"],
                message="message",
                message_group_id="import_applications_group",
                message_deduplication_id=str(uuid4()),  # ensures message uniqueness
            )
            assert message_id is not None

    def _check_is_data_available(self, count):
        response = self.task_executor.sqs_extended_client.receive_messages(
            queue_url=self.queue_response["QueueUrl"], max_number=1
        )
        assert len(response) == count


class AnyTaskExecutorService(TaskExecutorService):
    def message_executor(self, message):
        return message



from copy import deepcopy

import pytest
from fsd_utils import Decision
from fsd_utils import evaluate_response
from fsd_utils.decision.evaluate_response_against_schema import (
    _evaluate_with_supplied_operators,
)


TEST_SCHEMA_1 = {
    "aaa111": [
        {"answerValue": "charity", "result": Decision.PASS, "caveat": None},
        {
            "answerValue": "parish_council",
            "result": Decision.PASS,
            "caveat": None,
        },
        {
            "answerValue": "plc",
            "result": Decision.FAIL,
            "caveat": None,
        },
    ],
    "bbb222": [
        {"answerValue": True, "result": Decision.PASS, "caveat": None},
        {"answerValue": False, "result": Decision.FAIL, "caveat": None},
    ],
    "ccc333": [
        {"answerValue": "10", "result": Decision.PASS, "caveat": None},
        {
            "answerValue": "15",
            "result": Decision.PASS_WITH_CAVEATS,
            "caveat": "Try not to cut down trees: This is a bit high",
        },
        {"answerValue": "20", "result": Decision.FAIL, "caveat": None},
    ],
    "eee555": [
        {"answerValue": "a", "result": Decision.PASS, "caveat": None},
        {
            "answerValue": "b",
            "result": Decision.PASS_WITH_CAVEATS,
            "caveat": "Caveat heading: some caveat text",
        },
        {"answerValue": "c", "result": Decision.FAIL, "caveat": None},
    ],
    "fff666": [
        {
            "operator": "<=",
            "compareValue": 4,
            "result": Decision.PASS,
            "caveat": None,
        },
        {
            "operator": ">=",
            "compareValue": 5,
            "result": Decision.PASS_WITH_CAVEATS,
            "caveat": "A caveat: Try and reduce this",
        },
        {
            "operator": ">=",
            "compareValue": 10,
            "result": Decision.FAIL,
            "caveat": None,
        },
    ],
    # add a deliberatly weird set of conditions
    "ggg777": [
        {"answerValue": "10", "result": Decision.PASS, "caveat": None},
        {"answerValue": "7", "result": Decision.PASS, "caveat": None},
        {
            "operator": "<",
            "compareValue": 10,
            "result": Decision.PASS_WITH_CAVEATS,
            "caveat": "A caveat",
        },
        {
            "operator": ">",
            "compareValue": 10,
            "result": Decision.FAIL,
            "caveat": None,
        },
    ],
}


# Test forms that will pass all the conditions in TEST_SCHEMA_1
TEST_EOI_FORMS_1 = [
    {
        "name": "test-form-1",
        "questions": [
            {
                "category": "GMkooI",
                "fields": [
                    {
                        "answer": "charity",
                        "key": "aaa111",
                        "title": "What type of org are you?",
                        "type": "text",
                    },
                    {
                        "answer": "something",
                        "key": "ddd444",
                        "title": "Question with no bearing on decision",
                        "type": "text",
                    },
                ],
                "question": "Test Question 1",
                "status": "COMPLETED",
            }
        ],
        "status": "COMPLETED",
    },
    {
        "name": "test-form-2",
        "questions": [
            {
                "category": "GMkooI",
                "fields": [
                    {
                        "answer": True,
                        "key": "bbb222",
                        "title": "Do you have a plan in place",
                        "type": "list",
                    },
                    {
                        "answer": "10",
                        "key": "ccc333",
                        "title": "How many trees will you have to cut down?",
                        "type": "list",
                    },
                ],
                "question": "Test Question 2",
                "status": "COMPLETED",
            },
            {
                "category": "GMkooI",
                "fields": [
                    {
                        "answer": "a",
                        "key": "eee555",
                        "title": "Another question with caveats",
                        "type": "list",
                    },
                    {
                        "answer": "3",
                        "key": "fff666",
                        "title": "Pass or fail based on number",
                        "type": "NumberField",
                    },
                    {
                        "answer": "10",
                        "key": "ggg777",
                        "title": "Pass or fail based on number",
                        "type": "NumberField",
                    },
                ],
            },
        ],
        "status": "COMPLETED",
    },
]


@pytest.mark.parametrize(
    "answers,schema,exp_result,exp_caveats",
    [
        (  # All pass (no changes)
            {},
            TEST_SCHEMA_1,
            Decision.PASS,
            [],
        ),
        (  # All pass but one fail
            {"bbb222": False},
            TEST_SCHEMA_1,
            Decision.FAIL,
            [],
        ),
        (  # One pass with caveats, one fail
            {"ccc333": "15", "bbb222": False},
            TEST_SCHEMA_1,
            Decision.FAIL,
            [],
        ),
        (  # All pass but one pass with caveats
            {"ccc333": "15"},
            TEST_SCHEMA_1,
            Decision.PASS_WITH_CAVEATS,
            ["Try not to cut down trees: This is a bit high"],
        ),
        (  # Most pass but 2 pass with caveats - one number, one string, both value based
            {"eee555": "b", "ccc333": "15"},
            TEST_SCHEMA_1,
            Decision.PASS_WITH_CAVEATS,
            [
                "Try not to cut down trees: This is a bit high",
                "Caveat heading: some caveat text",
            ],
        ),
        (  # Most pass but 2 pass with caveats - one number - operator based, one string
            {"fff666": "6", "eee555": "b"},
            TEST_SCHEMA_1,
            Decision.PASS_WITH_CAVEATS,
            [
                "Caveat heading: some caveat text",
                "A caveat: Try and reduce this",
            ],
        ),
        (  # Most pass, one pass with caveats based on operator with contradicting value condition
            {"ggg777": "7"},
            TEST_SCHEMA_1,
            Decision.PASS_WITH_CAVEATS,
            [
                "A caveat",
            ],
        ),
        (  # Most pass, one pass with caveats based on operator
            {"ggg777": "3"},
            TEST_SCHEMA_1,
            Decision.PASS_WITH_CAVEATS,
            [
                "A caveat",
            ],
        ),
        (
            # most pass, one fails based on operator
            {"ggg777": "12"},
            TEST_SCHEMA_1,
            Decision.FAIL,
            [],
        ),
    ],
)
def test_schema_parsing(answers: dict, schema, exp_result, exp_caveats):
    # Create a copy of the test form and update with supplied answers for this scenario
    input_forms = deepcopy(TEST_EOI_FORMS_1)
    for answer in answers.items():
        for form in input_forms:
            for question in form["questions"]:
                for x in [
                    field for field in question["fields"] if field["key"] == answer[0]
                ]:
                    x["answer"] = answer[1]

    # get a result
    result = evaluate_response(schema, input_forms)

    # confirm result is as expected
    assert result["decision"] == exp_result
    assert result["caveats"] == exp_caveats


TEST_OPERATOR_CONDITIONS = [
    {
        "operator": "<=",
        "compareValue": 4,
        "result": Decision.PASS,
        "caveat": None,
    },
    {
        "operator": ">=",
        "compareValue": 5,
        "result": Decision.PASS_WITH_CAVEATS,
        "caveat": "A caveat: more than 5",
    },
    {
        "operator": "==",
        "compareValue": 7,
        "result": Decision.PASS_WITH_CAVEATS,
        "caveat": "A caveat: equals 7",
    },
    {
        "operator": ">",
        "compareValue": 10,
        "result": Decision.FAIL,
        "caveat": None,
    },
    {
        "operator": "<",
        "compareValue": 0,
        "result": Decision.PASS,
        "caveat": None,
    },
]


@pytest.mark.parametrize(
    "answer,exp_decision,exp_caveats",
    [
        (2, Decision.PASS, []),
        ("2", Decision.PASS, []),
        (-1, Decision.PASS, []),
        ("-1", Decision.PASS, []),
        (5, Decision.PASS_WITH_CAVEATS, ["A caveat: more than 5"]),
        ("5", Decision.PASS_WITH_CAVEATS, ["A caveat: more than 5"]),
        (
            7,
            Decision.PASS_WITH_CAVEATS,
            ["A caveat: more than 5", "A caveat: equals 7"],
        ),
        (
            "7",
            Decision.PASS_WITH_CAVEATS,
            ["A caveat: more than 5", "A caveat: equals 7"],
        ),
        (12, Decision.FAIL, []),
        ("12", Decision.FAIL, []),
    ],
)
def test_evaluate_operators(answer, exp_decision, exp_caveats):
    result = _evaluate_with_supplied_operators(TEST_OPERATOR_CONDITIONS, answer)
    assert result[0] == exp_decision
    assert result[1] == exp_caveats


@pytest.mark.parametrize(
    "operator",
    [
        ("<"),
        (">"),
        ("<="),
        (">="),
        ("=="),
    ],
)
def test_operator_validation_success(operator):
    condition = {
        "operator": operator,
        "compareValue": 4,
        "result": Decision.PASS,
        "caveat": None,
    }
    _evaluate_with_supplied_operators([condition], 1)
    # Should be no exception - will fail if one is thrown


@pytest.mark.parametrize(
    "operator",
    [
        ("<X"),
        (""),
        ("randomsomethingelse"),
        ("*"),
        ("and"),
        (None),
    ],
)
def test_operator_validation_failures(operator):
    condition = {
        "operator": operator,
        "compareValue": 4,
        "result": Decision.PASS,
        "caveat": None,
    }

    with pytest.raises(ValueError):
        _evaluate_with_supplied_operators([condition], 1)


def test_no_questions_hit_conditions():
    TEST_SCHEMA = {
        "does_not_exist": [
            {"answerValue": True, "result": Decision.PASS, "caveat": None},
            {"answerValue": False, "result": Decision.FAIL, "caveat": None},
        ],
    }
    result = evaluate_response(TEST_SCHEMA, TEST_EOI_FORMS_1)
    assert result["decision"] == Decision.PASS
    assert result["caveats"] == []


@pytest.mark.parametrize(
    "supplied_answer",
    [
        "a1",
        "--2",
        "-a3.356345",
        "400099.asdf",
        "£234",
        "£234234.00",
        "hello there",
        "45,6",
        "123£$%^&*()''`\"",
    ],
)
def test_answer_validation_failure(supplied_answer):
    condition = {
        "operator": "<",
        "compareValue": 4,
        "result": Decision.PASS,
        "caveat": None,
    }
    with pytest.raises(ValueError):
        _evaluate_with_supplied_operators([condition], supplied_answer)


@pytest.mark.parametrize(
    "supplied_answer",
    ["1", 1, "-2", "-3.356345", "400099.234234", 3434.5656, 0, "00000", 0000],
)
def test_answer_validation_success(supplied_answer):
    condition = {
        "operator": "<",
        "compareValue": 4,
        "result": Decision.PASS,
        "caveat": None,
    }
    _evaluate_with_supplied_operators([condition], supplied_answer)



from enum import IntEnum


class Decision(IntEnum):
    PASS = 0
    PASS_WITH_CAVEATS = 1
    FAIL = 2


VALID_OPERATORS = ["<", "<=", "==", ">=", ">"]


def _evaluate_with_supplied_operators(
    conditions_to_evaluate: list, supplied_answer: any
) -> tuple[Decision, list]:
    """Evaluates an expression built from the operator in the schmea, the value to compare, and the supplied answer.
    Uses the result of the evaluation to return a decision and applicable caveats
    Casts the value to an integer for comparison

    Args:
        conditions_to_evaluate (list): List of conditions from schema
        supplied_answer (any): Answer supplied to this question

    Raises:
        ValueError: If the operator from the schema is not supported, or if the supplied answer
        cannot be converted to a float

    Returns:
        tuple[Decision, list]: Tuple of the decision and the caveats (if there are any)
    """
    decision = Decision.PASS
    caveats = []
    for ec in conditions_to_evaluate:
        # validate supplied operator
        if ec["operator"] not in VALID_OPERATORS:
            raise ValueError(f"Operator {ec['operator']} is not supported")

        # validate that answer is numeric
        try:
            answer_as_number = float(supplied_answer)
        except ValueError:
            raise ValueError(
                f"Answer {supplied_answer} is not numeric so cannot be used with this condition: "
                f"[{ec['operator']} {ec['compareValue']}]"
            )

        # construct evaluation expression
        expression = f"answer {ec['operator']} value"

        # evaluation using supplied operator
        if eval(expression, {"answer": answer_as_number, "value": ec["compareValue"]}):
            # We met this condition
            decision = max(decision, ec["result"])
            if ec["caveat"]:
                caveats.append(ec["caveat"])

    if decision == Decision.FAIL:
        return decision, []  # don't return caveats for failure
    else:
        return decision, caveats


def evaluate_response(schema: dict, forms: dict) -> dict:
    """Takes in a decision schema and a set of forms containing responses, then makes a decision on the outcome

    Args:
        schema (dict): Schema defining decisions based on answers
        forms (dict): Response forms in form-builder json format

    Returns:
        dict: Results of decision:
            decision: value of Decision ENUM
            caveats: list of strings of caveats for answers if decision is 'Pass with caveats', otherwise empty list

    """
    result = {"decision": Decision.PASS, "caveats": []}

    # Loop through every form, then every response in that form
    for form in forms:
        for question in form["questions"]:
            for response in question["fields"]:
                questionId = response["key"]

                # Find out if there are any conditions that apply to this question
                if conditions := schema.get(questionId, None):

                    # Find out if there are any value-based conditions that match the response given for this question
                    if applicable_condition := next(
                        (
                            c
                            for c in conditions
                            if c.get("answerValue", None) == response["answer"]
                        ),
                        None,
                    ):
                        result["decision"] = max(
                            result["decision"], applicable_condition["result"]
                        )
                        if applicable_condition["caveat"]:
                            result["caveats"].append(applicable_condition["caveat"])

                    # Gather any operator-based conditions that apply to this question
                    if eval_conditions := [c for c in conditions if "operator" in c]:

                        # Got and get a decision based on this operator condition
                        decision, caveats = _evaluate_with_supplied_operators(
                            eval_conditions, response["answer"]
                        )

                        # Update result object with result of evaluating the operator-condition
                        result["decision"] = max(result["decision"], decision)
                        result["caveats"] += caveats

                    # If we failed on this question, we don't need to evaluate any further, just return a fail
                    if result["decision"] == Decision.FAIL:
                        return {"decision": Decision.FAIL, "caveats": []}

    return result



import json
from datetime import datetime

import boto3
from fsd_utils.services.aws_sqs_extended_client_exception import ExceptionMessages
from fsd_utils.services.aws_sqs_extended_client_exception import (
    SQSExtendedClientException,
)
from fsd_utils.services.aws_sqs_extended_client_util import check_message_attributes
from fsd_utils.services.aws_sqs_extended_client_util import get_s3_key
from fsd_utils.services.aws_sqs_extended_client_util import MESSAGE_POINTER_CLASS
from fsd_utils.services.aws_sqs_extended_client_util import RESERVED_ATTRIBUTE_NAME
from fsd_utils.services.aws_sqs_extended_client_util import validate_messages


class SQSExtendedClient:
    def __init__(
        self,
        aws_access_key_id=None,
        aws_secret_access_key=None,
        region_name="eu-west-2",
        endpoint_url=None,
        large_payload_support=None,
        always_through_s3=None,
        delete_payload_from_s3=None,
        logger=None,
        **kwargs,
    ):
        self.large_payload_support = large_payload_support
        self.always_through_s3 = always_through_s3
        self.delete_payload_from_s3 = delete_payload_from_s3
        self.logger = logger

        if aws_access_key_id and aws_secret_access_key:
            self.sqs_client = boto3.client(
                "sqs",
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                region_name=region_name,
                endpoint_url=endpoint_url,
                **kwargs,
            )
            self.s3_client = boto3.client(
                "s3",
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                region_name=region_name,
                endpoint_url=endpoint_url,
            )
        else:
            """
            if 'aws_access_key_id' and 'aws_access_key_id' are not provided make sure to provide
            'AWS_ACCESS_KEY_ID' and 'AWS_SECRET_ACCESS_KEY' with environment variables
            """
            self.sqs_client = boto3.client(
                "sqs",
                region_name=region_name,
                endpoint_url=endpoint_url,
                **kwargs,
            )
            self.s3_client = boto3.client(
                "s3",
                region_name=region_name,
                endpoint_url=endpoint_url,
            )

    def get_queues(self, prefix=None):
        """
        Gets a list of SQS queues. When a prefix is specified, only queues with names
        that start with the prefix are returned.

        :param prefix: The prefix used to restrict the list of returned queues.
        :return: A list of Queue names.
        """
        if prefix:
            return self._get_queue_names(
                self.sqs_client.list_queues(QueueNamePrefix=prefix)["QueueUrls"]
            )
        return self._get_queue_names(self.sqs_client.list_queues()["QueueUrls"])

    def submit_single_message(
        self,
        queue_url,
        message,
        extra_attributes: dict = None,
        message_group_id=None,
        message_deduplication_id=None,
    ):
        sqs_message_attributes = {
            "message_created_at": {
                "StringValue": str(datetime.now()),
                "DataType": "String",
            },
        }
        message_body, message_attributes = self._store_message_in_s3(
            message, sqs_message_attributes, extra_attributes
        )
        # add extra message attributes (if provided)
        if extra_attributes:
            for key, value in extra_attributes.items():
                message_attributes[key] = value

        response = self.sqs_client.send_message(
            QueueUrl=queue_url,
            MessageBody=message_body,
            MessageAttributes=message_attributes,
            MessageGroupId=message_group_id,
            MessageDeduplicationId=message_deduplication_id,
        )
        # Check if the delete operation succeeded, if not raise an error?
        status_code = response["ResponseMetadata"]["HTTPStatusCode"]
        if status_code != 200:
            raise SQSExtendedClientException(
                ExceptionMessages.FAILED_SUBMIT_MESSAGE.format(status_code)
            )
        message_id = response["MessageId"]
        self.logger.info(f"Called SQS and submitted the message and id [{message_id}]")
        return message_id

    def receive_messages(self, queue_url, max_number, visibility_time=1, wait_time=1):
        """
        Receive a batch of messages in a single request from an SQS queue.
        :param queue_url: SQS Queue url
        :param max_number: The maximum number of messages to receive. The actual number
                        of messages received might be less.
        :param visibility_time: The maximum time for message to temporarily invisible to other receivers.
                                This gives the initial receiver a chance to process the message. If the receiver
                                successfully processes and deletes the message within the visibility timeout,
                                the message is removed from the queue.
        :param wait_time: The maximum time to wait (in seconds) before returning. When
                        this number is greater than zero, long polling is used. This
                        can result in reduced costs and fewer false empty responses.
        :return: The list of Message objects received. These each contain the body
                of the message and metadata and custom attributes.
        """
        response = self.sqs_client.receive_message(
            QueueUrl=queue_url,
            AttributeNames=["SentTimestamp", "ApproximateReceiveCount"],
            MessageAttributeNames=["All"],
            MaxNumberOfMessages=max_number,
            VisibilityTimeout=visibility_time,
            WaitTimeSeconds=wait_time,
        )

        # Check if the delete operation succeeded, if not raise an error?
        status_code = response["ResponseMetadata"]["HTTPStatusCode"]
        if status_code != 200:
            raise SQSExtendedClientException(
                ExceptionMessages.FAILED_RECEIVE_MESSAGE.format(status_code)
            )
        messages = []
        if "Messages" in response.keys():
            messages = response["Messages"]
            self.logger.info(f"Called SQS and received [{len(messages)}] messages")
        elif response["ResponseMetadata"]["HTTPStatusCode"] == 200:
            return []

        if self.large_payload_support and self.always_through_s3:
            extended_message = []
            for msg in messages:
                dict_msg = {
                    "sqs": msg,
                    "s3": self._retrieve_message_from_s3(msg["Body"]),
                }
                extended_message.append(dict_msg)
            return extended_message
        return messages

    def delete_messages(self, queue_url, messages):
        validate_messages(messages)
        reciept_handles_to_delete = []
        if self.large_payload_support and self.always_through_s3:
            return self._delete_msg_from_sqs_and_s3(
                messages, queue_url, reciept_handles_to_delete
            )
        return self._delete_msg_from_sqs(messages, queue_url, reciept_handles_to_delete)

    def _delete_msg_from_sqs_and_s3(
        self, messages, queue_url, reciept_handles_to_delete
    ):
        self._delete_message_from_s3(messages, reciept_handles_to_delete)
        entries = [
            {"Id": str(ind), "ReceiptHandle": receipt_handle}
            for ind, receipt_handle in enumerate(reciept_handles_to_delete)
        ]
        response = self.sqs_client.delete_message_batch(
            QueueUrl=queue_url, Entries=entries
        )
        # Check if the delete operation succeeded, if not raise an error?
        status_code = response["ResponseMetadata"]["HTTPStatusCode"]
        if status_code != 200:
            raise SQSExtendedClientException(
                ExceptionMessages.FAILED_DELETE_MESSAGE.format(status_code)
            )
        self.logger.info("Called SQS and deleted the message")
        if "Successful" in response:
            for msg_meta in response["Successful"]:
                self.logger.info(
                    f"Deleted {reciept_handles_to_delete[int(msg_meta['Id'])]}"
                )
        if "Failed" in response:
            for msg_meta in response["Failed"]:
                self.logger.info(
                    f"Could not delete {reciept_handles_to_delete[int(msg_meta['Id'])]}"
                )
        return response

    def _delete_msg_from_sqs(self, messages, queue_url, reciept_handles_to_delete):
        for message in messages:
            reciept_handles_to_delete.append(message["ReceiptHandle"])
        entries = [
            {"Id": str(ind), "ReceiptHandle": receipt_handle}
            for ind, receipt_handle in enumerate(reciept_handles_to_delete)
        ]
        response = self.sqs_client.delete_message_batch(
            QueueUrl=queue_url, Entries=entries
        )
        # Check if the delete operation succeeded, if not raise an error?
        status_code = response["ResponseMetadata"]["HTTPStatusCode"]
        if status_code != 200:
            raise SQSExtendedClientException(
                ExceptionMessages.FAILED_DELETE_MESSAGE.format(status_code)
            )
        self.logger.info("Called SQS and deleted the message")
        if "Successful" in response:
            for msg_meta in response["Successful"]:
                self.logger.info(
                    f"Deleted {reciept_handles_to_delete[int(msg_meta['Id'])]}"
                )
        if "Failed" in response:
            for msg_meta in response["Failed"]:
                self.logger.info(
                    f"Could not delete {reciept_handles_to_delete[int(msg_meta['Id'])]}"
                )
        return response

    def _delete_message_from_s3(self, messages, reciept_handles_to_delete):
        for msg in messages:
            message_body = json.loads(msg["Body"])
            reciept_handles_to_delete.append(msg["ReceiptHandle"])
            if not (
                isinstance(message_body, list)
                and len(message_body) == 2
                and isinstance(message_body[1], dict)
            ):
                raise SQSExtendedClientException(
                    ExceptionMessages.INVALID_FORMAT_WHEN_RETRIEVING_STORED_S3_MESSAGES
                )
            s3_details = message_body[1]
            s3_bucket_name, s3_key = s3_details["s3BucketName"], s3_details["s3Key"]

            response = self.s3_client.delete_object(Bucket=s3_bucket_name, Key=s3_key)
            # Check if the delete operation succeeded, if not raise an error?
            status_code = response["ResponseMetadata"]["HTTPStatusCode"]
            if status_code != 204:
                raise SQSExtendedClientException(
                    ExceptionMessages.FAILED_DELETE_MESSAGE.format(status_code)
                )
            self.logger.info("Called S3 and deleted the message")

    def _retrieve_message_from_s3(self, message_body: str) -> str:
        """
        Responsible for retrieving a message payload from a S3 Bucket, if it exists
        :message_body: A string containing the first element to be the S3 class pointer
        and the second element to be a dictionary consisting of the s3BucketName and
        the s3Key for the bucket.
        """
        message_body = json.loads(message_body)
        if not (
            isinstance(message_body, list)
            and len(message_body) == 2
            and isinstance(message_body[1], dict)
        ):
            raise SQSExtendedClientException(
                ExceptionMessages.INVALID_FORMAT_WHEN_RETRIEVING_STORED_S3_MESSAGES
            )
        s3_details = message_body[1]
        s3_bucket_name, s3_key = s3_details["s3BucketName"], s3_details["s3Key"]
        response = self.s3_client.get_object(Bucket=s3_bucket_name, Key=s3_key)
        # The message body is under a wrapper class called StreamingBody
        status_code = response["ResponseMetadata"]["HTTPStatusCode"]
        if status_code != 200:
            raise SQSExtendedClientException(
                ExceptionMessages.FAILED_RECEIVE_MESSAGE.format(status_code)
            )
        self.logger.info("Called S3 and received the message")
        streaming_body = response["Body"]
        message_body = streaming_body.read().decode()
        return message_body

    def _store_message_in_s3(
        self, message_body: str, message_attributes: dict, extra_attributes: dict
    ) -> (str, dict):
        """
        Responsible for storing a message payload in a S3 Bucket
        :message_body: A UTF-8 encoded version of the message body
        :message_attributes: A dictionary consisting of message attributes
        :extra_attributes: A dictionary consisting of message attributes
        Each message attribute consists of the name (key) along with a
        type and value of the message body. The following types are supported
        for message attributes: StringValue, BinaryValue and DataType.
        """
        if len(message_body) == 0:
            # Message cannot be empty
            raise SQSExtendedClientException(ExceptionMessages.INVALID_MESSAGE_BODY)

        if self.large_payload_support and self.always_through_s3:
            # Check message attributes for ExtendedClient related constraints
            check_message_attributes(message_attributes)
            encoded_body = message_body.encode("utf-8")

            # Modifying the message attributes for storing it in the Queue
            message_attributes[RESERVED_ATTRIBUTE_NAME] = {}
            attribute_value = {
                "DataType": "Number",
                "StringValue": str(len(encoded_body)),
            }
            message_attributes[RESERVED_ATTRIBUTE_NAME] = attribute_value

            # S3 Key should either be a constant or be a random uuid4 string.
            s3_key = get_s3_key(message_attributes, extra_attributes)

            # Adding the object into the bucket
            response = self.s3_client.put_object(
                Body=encoded_body, Bucket=self.large_payload_support, Key=s3_key
            )
            # Check if the delete operation succeeded, if not raise an error?
            status_code = response["ResponseMetadata"]["HTTPStatusCode"]
            if status_code != 200:
                raise SQSExtendedClientException(
                    ExceptionMessages.FAILED_SUBMIT_MESSAGE.format(status_code)
                )
            self.logger.info(
                f"Message added to S3 bucket [{self.large_payload_support}]"
            )
            # Modifying the message body for storing it in the Queue
            message_body = json.dumps(
                [
                    MESSAGE_POINTER_CLASS,
                    {"s3BucketName": self.large_payload_support, "s3Key": s3_key},
                ]
            )
        return message_body, message_attributes

    def _get_queue_names(self, queues):
        if queues:
            self.logger.info(f"Got queues: {', '.join([q for q in queues])}")
            queue_names = [url.split("/")[-1] for url in queues]
            return queue_names
        self.logger.info("No queues found.")
        return []



import pytest
from fsd_utils.mapping.application.application_utils import convert_bool_value
from fsd_utils.mapping.application.application_utils import format_answer
from fsd_utils.mapping.application.application_utils import format_checkbox
from fsd_utils.mapping.application.application_utils import format_date_month_year
from fsd_utils.mapping.application.application_utils import format_month_year
from fsd_utils.mapping.application.application_utils import format_radio_field
from fsd_utils.mapping.application.application_utils import simplify_title
from fsd_utils.mapping.application.free_text import FreeText
from fsd_utils.mapping.application.languages import CY
from fsd_utils.mapping.application.languages import EN
from fsd_utils.mapping.application.multi_input import MultiInput
from fsd_utils.mapping.application.multi_input_utils import ProcessTypes
from fsd_utils.mapping.application.qa_mapping import extract_questions_and_answers
from tests.test_data_utils import multi_input_test_data
from tests.test_data_utils import test_data_sort_questions_answers
from tests.test_data_utils import test_data_sort_questions_answers_welsh


@pytest.mark.parametrize(
    "input_data, language, expected_response",
    [
        (True, "en", "Yes"),
        (True, None, "Yes"),
        (True, "", "Yes"),
        (True, "cy", "Oes"),
        (False, "en", "No"),
        (False, None, "No"),
        (False, "", "No"),
        (False, "cy", "Nac Oes"),
        (["address", True], "en", ["address", "Yes"]),
        (["address", True], None, ["address", "Yes"]),
        (["address", True], "", ["address", "Yes"]),
        (["address", True], "cy", ["address", "Oes"]),
        (["address", False], "en", ["address", "No"]),
        (["address", False], None, ["address", "No"]),
        (["address", False], "", ["address", "No"]),
        (["address", False], "cy", ["address", "Nac Oes"]),
        (None, "en", "Not provided"),
        (None, None, "Not provided"),
        (None, "", "Not provided"),
        (None, "cy", "Heb ei ddarparu"),
        ([["name", True], "en", ["name", "Yes"]]),
        ([["name", True], None, ["name", "Yes"]]),
        ([["name", True], "", ["name", "Yes"]]),
        ([["name", True], "cy", ["name", "Oes"]]),
        ([["name", False], "en", ["name", "No"]]),
        ([["name", False], None, ["name", "No"]]),
        ([["name", False], "", ["name", "No"]]),
        ([["name", False], "cy", ["name", "Nac Oes"]]),
    ],
)
def test_convert_bool_values(input_data, expected_response, language):
    response = convert_bool_value(input_data, language)
    assert response == expected_response


@pytest.mark.parametrize(
    "input_data, expected_response",
    [
        ("null", ""),
        (None, "Not provided"),
        (
            "555 dummy road, null, beautiful town, optional county, UU8 8UU",
            "555 dummy road, beautiful town, optional county, UU8 8UU",
        ),
        ("https://example.com/test-the-link", "https://example.com/test-the-link"),
        (["four five six"], ["four five six"]),
        ("ten eleven twelve", "ten eleven twelve"),
        ({"title": ["one-two-three"]}, {"title": ["one-two-three"]}),
    ],
)
def test_format_answer(input_data, expected_response):
    response = format_answer(input_data, EN)
    assert response == expected_response


@pytest.mark.parametrize(
    "section_name, remove_text, expected_response",
    [
        (
            "organisation-type-cof-round",
            ["cof", "ns"],
            ["organisation", "type"],
        ),
        (
            "organisation-name-ns-round",
            ["cof", "ns"],
            ["organisation", "name"],
        ),
        (
            "organisation-name-dpi-round",
            ["cof", "dpi"],
            ["organisation", "name"],
        ),
        (
            "organisation-name-cyp",
            ["cof", "cyp"],
            ["organisation", "name"],
        ),
        (
            "total-expected-cost-hsra",
            ["cof", "cyp", "hsra"],
            ["total", "expected", "cost"],
        ),
        ("organisation-address", ["cof", "ns"], ["organisation", "address"]),
    ],
)
def test_simplify_title(section_name, remove_text, expected_response):
    response = simplify_title(section_name, remove_text)
    assert response == expected_response


@pytest.mark.parametrize(
    "input_value, expected_response",
    [
        ("<p> Hello world </p>", "Hello world"),
        ("<ul><li>Item 1</li><li>Item 2</li></ul>", ". Item 1\n     . Item 2"),
        (
            "<ol><li>Item 1</li><li>Item 2</li></ol>",
            "1. Item 1\n     2. Item 2",
        ),
        ("text without html tags", "text without html tags"),
        (None, None),
        (["one", "two"], ["one", "two"]),
        (True, True),
        ("https://my-website.com", "https://my-website.com"),
    ],
)
def test_remove_html_tags(input_value, expected_response):

    response = FreeText.remove_html_tags(input_value)
    assert response == expected_response


def test_extract_questions_and_answers(app_context):
    forms = test_data_sort_questions_answers["forms"]

    response = extract_questions_and_answers(forms)
    assert response == test_data_sort_questions_answers["questions_answers"]


def test_extract_questions_and_answers_welsh(app_context):
    forms = test_data_sort_questions_answers_welsh["forms"]

    response = extract_questions_and_answers(forms, CY)
    assert response == test_data_sort_questions_answers_welsh["questions_answers"]


def test_extract_questions_and_answers_fail(app_context):
    forms = test_data_sort_questions_answers["incorrect_form_data"]

    with pytest.raises(Exception) as exc:
        extract_questions_and_answers(forms)

    assert str(exc.value) == test_data_sort_questions_answers["exception_message"]


@pytest.mark.parametrize(
    "input_value, expected_response",
    [
        (
            [
                "health-interventions",
                "employment-support",
            ],
            ". Health interventions\n     . Employment support",
        ),
        (
            ["Survivors of domestic abuse", "ethnic minorities"],
            ". Survivors of domestic abuse\n     . Ethnic minorities",
        ),
        (
            ["gis (geographical-information-system)"],
            ". Gis (geographical information system)",
        ),
    ],
)
def test_format_checkbox(input_value, expected_response):

    response = format_checkbox(input_value)
    assert response == expected_response


@pytest.mark.parametrize(
    "input_value, expected_response",
    [
        ("health-interventions", "Health interventions"),
        ("https://my-website.com", "https://my-website.com"),
        (None, None),
        (True, True),
        (["test-list-answer"], ["test-list-answer"]),
        (
            "gis (geographical-information-system)",
            "Gis (geographical information system)",
        ),
        ("capital", "Capital"),
    ],
)
def test_format_radio_field(input_value, expected_response):

    response = format_radio_field(input_value)
    assert response == expected_response


@pytest.mark.parametrize(
    "input_value, expected_response",
    [
        ("2023-11", "November 2023"),
        ("11-2023", "November 2023"),
    ],
)
def test_format_month_year(input_value, expected_response):

    response = format_month_year(input_value)
    assert response == expected_response


@pytest.mark.parametrize(
    "input_value, expected_response",
    [
        ("2023-11-01", "01 November 2023"),
        ("2023-11-1", "01 November 2023"),
        ("01-11-2023", "01 November 2023"),
        ("1-11-2023", "01 November 2023"),
        ("2023-11-19", "19 November 2023"),
        ("2023-1-1", "01 January 2023"),
    ],
)
def test_format_date_month_year(input_value, expected_response):

    response = format_date_month_year(input_value)
    assert response == expected_response


class TestMultiInput:
    @pytest.mark.parametrize(
        "input_data, expected_response",
        [
            (
                multi_input_test_data["format_data"]["multiple_values"]["input_data"],
                multi_input_test_data["format_data"]["multiple_values"][
                    "expected_response"
                ],
            ),
            (
                multi_input_test_data["format_data"]["single_value"]["input_data"],
                multi_input_test_data["format_data"]["single_value"][
                    "expected_response"
                ],
            ),
        ],
    )
    def test_format_data(self, app_context, input_data, expected_response):
        response = MultiInput.format_data(input_data, EN)
        assert response == expected_response

    @pytest.mark.parametrize(
        "input_data, expected_response",
        [
            (
                multi_input_test_data["map_data"]["dict_two_str_values"]["input_data"],
                multi_input_test_data["map_data"]["dict_two_str_values"][
                    "expected_response"
                ],
            ),
            (
                multi_input_test_data["map_data"]["dict_int_value"]["input_data"],
                multi_input_test_data["map_data"]["dict_int_value"][
                    "expected_response"
                ],
            ),
            (
                multi_input_test_data["map_data"]["dict_none_values"]["input_data"],
                multi_input_test_data["map_data"]["dict_none_values"][
                    "expected_response"
                ],
            ),
        ],
    )
    def test_map_multi_input_data(self, app_context, input_data, expected_response):
        response = MultiInput.map_multi_input_data(input_data, EN)
        assert response == expected_response

    @pytest.mark.parametrize(
        "input_data, expected_response",
        [
            (
                {"hgYuhh": "Sky builders"},
                {"6ba7b810-9dad-11d1-80b4-00c04fd430c8": "Sky builders"},
            )
        ],
    )
    def test_str_single_item(
        self, app_context, input_data, expected_response, mock_uuid4
    ):
        response = ProcessTypes._str_single_item(input_data, sorted_data={})
        assert response == expected_response

    @pytest.mark.parametrize(
        "input_data, expected_response",
        [
            (
                {
                    "PrulfI": {"PrulfI__month": 6, "PrulfI__year": 2018},
                },
                {"date_6ba7b810-9dad-11d1-80b4-00c04fd430c8": "June 2018"},
            )
        ],
    )
    def test_dict_single_item(
        self, app_context, input_data, expected_response, mock_uuid4
    ):
        response = ProcessTypes._dict_single_item(input_data, sorted_data={})
        assert response == expected_response

    @pytest.mark.parametrize(
        "input_data, expected_response",
        [
            (
                {
                    "addressLine1": "123 Main St",
                    "addressLine2": "Apt 4B",
                    "county": "Example County",
                    "postcode": "12345",
                    "town": "Sample Town",
                    "extra_field": "This should be ignored",
                },
                "123 Main St, Apt 4B, Example County, 12345, Sample Town",
            )
        ],
    )
    def test_validate_address(
        self, app_context, input_data, expected_response, mock_uuid4
    ):
        response = ProcessTypes.validate_address(input_data)
        assert response == expected_response

    @pytest.mark.parametrize(
        "input_data, expected_response",
        [
            ({"key1": "Value One", "key2": "Value Two"}, ["Value One", "Value Two"]),
            (
                {
                    "HpLJDu": "Milestone one",
                    "LZbOBu": {"LZbOBu__month": 6, "LZbOBu__year": 2024},
                },
                ["Milestone one", "June 2024"],
            ),
        ],
    )
    def test_value_len_two(self, app_context, input_data, expected_response):
        response = ProcessTypes.value_len_two(input_data)
        assert response == expected_response

    @pytest.mark.parametrize(
        "input_data, expected_response",
        [
            (
                {"key1": "Value One", "key2": 42, "key3": "Value Three"},
                ([42, "Value Three"], "Value One"),
            ),
            (
                {
                    "gLqiyJ": "Capital title",
                    "yuzbjT": 200,
                    "HpLJyL": {"HpLJyL__month": 3, "HpLJyL__year": 2023},
                    "MadvIr": "Capital",
                },
                ((["March 2023", 200, "Capital"], "Capital title")),
            ),
        ],
    )
    def test_value_len_greater_two(self, app_context, input_data, expected_response):
        response = ProcessTypes.value_len_greater_two(input_data)
        assert response == expected_response

    @pytest.mark.parametrize(
        "input_data, expected_response",
        [
            (
                {"hjgthY__date": 11, "hjgthY__month": 2, "hjgthY__year": 2023},
                ("11 February 2023"),
            ),
        ],
    )
    def test_validated_iso_value(self, app_context, input_data, expected_response):
        response = ProcessTypes.validated_iso_value(input_data)
        assert response == expected_response

    @pytest.mark.parametrize(
        "input_data, expected_response",
        [
            (
                "date_6ba7b810-9dad-11d1-80b4-00c04fd430c8",
                True,
            ),
            (
                "6ba7b810-9dad-11d1-80b4-00c04fd430c8",
                True,
            ),
            (
                "date_6ba7b810-9dad-11d1",
                False,
            ),
        ],
    )
    def test_is_valid_uuid(self, app_context, input_data, expected_response):
        response = ProcessTypes.is_valid_uuid(input_data)
        assert response == expected_response



import logging
import os


class CommonConfig:

    FSD_LOG_LEVELS = {
        "development": logging.DEBUG,
        "unit_test": logging.DEBUG,
        "dev": logging.INFO,
        "test": logging.WARN,
        "production": logging.ERROR,
    }

    # ---------------
    #  General App Config
    # ---------------
    SECRET_KEY = os.getenv("SECRET_KEY", "secret_key")
    SESSION_COOKIE_NAME = os.getenv("SESSION_COOKIE_NAME", "session_cookie")
    FLASK_ENV = os.getenv("FLASK_ENV")
    if not FLASK_ENV:
        raise KeyError("FLASK_ENV is not present in environment")
    try:
        FSD_LOG_LEVEL = FSD_LOG_LEVELS["FLASK_ENV"]
    except KeyError:
        FSD_LOG_LEVEL = FSD_LOG_LEVELS["production"]

    # ---------------
    # Dummy Hosts
    # ---------------

    TEST_ACCOUNT_STORE_API_HOST = "account_store"
    TEST_APPLICATION_STORE_API_HOST = "application_store"
    TEST_ASSESSMENT_STORE_API_HOST = "assessment_store"
    TEST_FUND_STORE_API_HOST = "fund_store"
    TEST_NOTIFICATION_SERVICE_HOST = "notification_service"

    # ---------------
    #  Account hosts, endpoints
    # ---------------

    ACCOUNT_STORE_API_HOST = os.getenv(
        "ACCOUNT_STORE_API_HOST", TEST_ACCOUNT_STORE_API_HOST
    )
    ACCOUNTS_ENDPOINT = "/accounts"
    ACCOUNT_ENDPOINT = "/accounts/{account_id}"

    # ---------------
    #  Application hosts, endpoints
    # ---------------

    APPLICATION_STORE_API_HOST = os.getenv(
        "APPLICATION_STORE_API_HOST", TEST_APPLICATION_STORE_API_HOST
    )
    APPLICATIONS_ENDPOINT = "/applications"
    APPLICATION_ENDPOINT = "/applications/{application_id}"
    APPLICATION_STATUS_ENDPOINT = "/applications/{application_id}/status"
    APPLICATION_SEARCH_ENDPOINT = "/applications?{params}"

    # ---------------
    # Assessment hosts, endpoints
    # ---------------

    ASSESSMENT_STORE_API_HOST = os.getenv(
        "ASSESSMENT_STORE_API_HOST", TEST_ASSESSMENT_STORE_API_HOST
    )

    # ---------------
    #  Fund hosts, endpoints
    # ---------------

    FUND_STORE_API_HOST = os.getenv("FUND_STORE_API_HOST", TEST_FUND_STORE_API_HOST)
    FUNDS_ENDPOINT = "/funds"
    FUND_ENDPOINT = "/funds/{fund_id}"  # account id in assessment store

    ROUNDS_ENDPOINT = "/funds/{fund_id}/rounds"
    ROUND_ENDPOINT = "/funds/{fund_id}/rounds/{round_id}"

    # ---------------
    #  Notification hosts, endpoints, fields
    # ---------------

    NOTIFICATION_SERVICE_HOST = os.getenv("NOTIFICATION_SERVICE_HOST")
    NOTIFICATION_SEND_ENDPOINT = "/send"

    # ---------------
    #  SESSION TIMELIMITS
    # ---------------

    FSD_SESSION_TIMEOUT_SECONDS = 86400  # (24hrs)
    WTF_CSRF_TIME_LIMIT = FSD_SESSION_TIMEOUT_SECONDS + 10  # (buffer secconds)

    # ---------------
    #  Talisman Settings
    # ---------------

    # Allow inline scripts for swagger docs (for Talisman Config)
    SWAGGER_CSP = {
        "script-src": ["'self'", "'unsafe-inline'"],
        "style-src": ["'self'", "'unsafe-inline'"],
    }

    # Content Security Policy
    SECURE_CSP = {
        "default-src": "'self'",
        "script-src": [
            "'self'",
            "'sha256-+6WnXIl4mbFTCARd8N3COQmT3bJJmo32N8q8ZSQAIcU='",
            "'sha256-l1eTVSK8DTnK8+yloud7wZUqFrI0atVo6VlC6PJvYaQ='",
            "'sha256-864+qKuyx4sP6oZXnPPGzxGuBEHMZ132TQ4RPp+zn44='",
            "'sha256-Z4Ew9hvNfnmh80vfWPHh13YzZdj0gTwoY6COVqIdaV0='",
            "'sha256-z+p4q2n8BOpGMK2/OMOXrTYmjbeEhWQQHC3SF/uMOyg='",
            "'sha256-RgdCrr7A9yqYVstE6QiM/9RNRj4bYipcUa2C2ywQT1A='",
            "https://tagmanager.google.com",
            "https://www.googletagmanager.com",
            "https://*.google-analytics.com",
        ],
        "connect-src": [
            "'self'",
            "https://*.google-analytics.com",
        ],  # APPLICATION_STORE_API_HOST_PUBLIC,
        "img-src": ["data:", "'self'", "https://ssl.gstatic.com"],
    }

    # Security headers and other policies
    FSD_REFERRER_POLICY = "strict-origin-when-cross-origin"
    FSD_SESSION_COOKIE_SAMESITE = "Lax"
    FSD_PERMISSIONS_POLICY = {"interest-cohort": "()"}
    FSD_DOCUMENT_POLICY = {}
    FSD_FEATURE_POLICY = {
        "microphone": "'none'",
        "camera": "'none'",
        "geolocation": "'none'",
    }

    DENY = "DENY"
    SAMEORIGIN = "SAMEORIGIN"
    ALLOW_FROM = "ALLOW-FROM"
    ONE_YEAR_IN_SECS = 31556926

    FORCE_HTTPS = False

    TALISMAN_SETTINGS = {
        "feature_policy": FSD_FEATURE_POLICY,
        "permissions_policy": FSD_PERMISSIONS_POLICY,
        "document_policy": FSD_DOCUMENT_POLICY,
        "force_https": FORCE_HTTPS,
        "force_https_permanent": False,
        "force_file_save": False,
        "frame_options": "SAMEORIGIN",
        "frame_options_allow_from": None,
        "strict_transport_security": True,
        "strict_transport_security_preload": True,
        "strict_transport_security_max_age": ONE_YEAR_IN_SECS,
        "strict_transport_security_include_subdomains": True,
        "content_security_policy": SECURE_CSP,
        "content_security_policy_report_uri": None,
        "content_security_policy_report_only": False,
        "content_security_policy_nonce_in": None,
        "referrer_policy": FSD_REFERRER_POLICY,
        "session_cookie_secure": True,
        "session_cookie_http_only": True,
        "session_cookie_samesite": FSD_SESSION_COOKIE_SAMESITE,
        "x_content_type_options": True,
        "x_xss_protection": True,
    }

    FSD_LANG_COOKIE_NAME = "language"

    # ---------------
    #  Feature Toggles
    # ---------------

    dev_feature_configuration = {"MULTIFUND_DASHBOARD": True}

    prod_feature_configuration = {"MULTIFUND_DASHBOARD": True}



# flake8: noqa: E501
import pytest
from fsd_utils.simple_utils.data_utils import get_remote_data_as_json
from requests.exceptions import HTTPError


class TestDataUtils:
    def test_get_remote_data_json(self, flask_test_client):
        postcode = "SW1P 4DF"
        result = get_remote_data_as_json(f"https://postcodes.io/postcodes/{postcode}")
        assert postcode == result["result"]["postcode"]

    def test_get_remote_data_json_404(self, flask_test_client):
        with pytest.raises(HTTPError):
            get_remote_data_as_json("https://postcodes.io/postcodes/QQ11QQ")


multi_input_test_data = {
    "format_data": {
        "multiple_values": {
            "input_data": {
                "trusts one": [
                    125,
                    "1 April 2023 to 31 March 2024",
                    "Capital",
                    True,
                ],
                "trust two": [
                    456,
                    "1 April 2024 to 31 March 2025",
                    "Revenue",
                    False,
                ],
            },
            "expected_response": [
                "trusts one \n      . 125\n      . 1 April 2023 to 31 March 2024\n      . Capital\n      . Yes",
                "\n     trust two \n      . 456\n      . 1 April 2024 to 31 March 2025\n      . Revenue\n"
                "      . No",
            ],
        },
        "single_value": {
            "input_data": {
                "bbd0ec2a-972f-4d06-bf93-bf24786c3859": "Sky builders",
                "ac8bbdfe-6a39-45b8-8c0a-6558148388d1": "trust builders",
            },
            "expected_response": [". Sky builders", "\n     . trust builders"],
        },
    },
    "map_data": {
        "multiple_values": {
            "input_data": [
                {
                    "AfAKxk": "trusts one",
                    "CrcLtW": 125,
                    "ndySbC": "1 April 2023 to 31 March 2024",
                    "pATWyM": "Capital",
                    "sIFBGc": True,
                },
                {
                    "AfAKxk": "trust two",
                    "CrcLtW": 456,
                    "ndySbC": "1 April 2024 to 31 March 2025",
                    "pATWyM": "Revenue",
                    "sIFBGc": False,
                },
            ],
            "expected_response": "* trusts one \n       . 125\n       . 1 April 2023 to 31 March 2024\n       . Capital\n       . Yes\n\n     * trust two \n       . 456\n       . 1 April 2024 to 31 March 2025\n       . Revenue\n       . No",  # noqa
        },
        "single_value": {
            "input_data": [
                {"CZZYvE": "Sky builders"},
                {"CZZYvE": "trust builders"},
            ],
            "expected_response": ". Sky builders\n     . trust builders",
        },
        "integer_values": {
            "input_data": [
                {"GLQlOh": "cost one", "JtwkMy": 4444},
                {"GLQl6y": "cost two", "JtwkMt": 4455},
            ],
            "expected_response": "* cost one \n       . 4444\n\n     * cost two \n       . 4455",
        },
        "dict_two_str_values": {
            "input_data": [
                {"CZZvN": "Name", "jhgvDjv": "Xose"},
            ],
            "expected_response": (". Name: Xose"),
        },
        "dict_int_value": {
            "input_data": [
                {"hGsUaZ": "Revenue Costs", "UyaAHw": 817},
            ],
            "expected_response": (". Revenue Costs: 817"),
        },
        "dict_none_values": {
            "input_data": [
                {
                    "JizgZP": 0,
                    "gLQlyJ": "This contribute £30000",
                    "kjuHtl": "1 April 2024 to 31 March 2025",
                }
            ],
            "expected_response": "This contribute £30000 \n      . 0\n      . 1 April 2024 to 31 March 2025",
        },
    },
}


test_data_sort_questions_answers = {
    "forms": [
        {
            "name": "funding-required-ns",
            "questions": [
                {
                    "fields": [
                        {
                            "answer": "both-revenue-and-capital",
                            "key": "NxVqXd",
                            "title": "What funding are you applying for?",
                            "type": "list",
                        },
                        {
                            "key": "NxVqXd",
                            "title": "What funding are you applying?",
                            "type": "list",
                            "answer": "capital",
                        },
                        {
                            "key": "tZoOKx",
                            "title": "Partner organisation details",
                            "type": "multiInput",
                            "answer": [
                                {
                                    "GpLJDu": "version1",
                                    "IXjMWp": {
                                        "addressLine1": "La la land",
                                        "addressLine2": "",
                                        "town": "Mars",
                                        "county": "",
                                        "postcode": "XA15 1AX",
                                    },
                                    "MKbOlA": "https://www.wikipedia.org/",
                                    "OghGGr": None,
                                    "RphKTp": None,
                                }
                            ],
                        },
                        {
                            "key": "tZoOKL",
                            "title": "Partner's Job",
                            "type": "multiInput",
                            "answer": [
                                {
                                    "GpLJDu": "version2",
                                    "IXjMWp": {
                                        "addressLine1": "La la la land",
                                        "addressLine2": "",
                                        "town": "Moon",
                                        "county": "",
                                        "postcode": "XA15 1AL",
                                    },
                                }
                            ],
                        },
                        {
                            "key": "NxSxCd",
                            "title": "Do you want to mark this section as complete?",
                            "type": "boolean",
                            "answer": "markAsComplete",
                        },
                    ],
                    "question": "What funding are you applying for?",
                },
                {
                    "fields": [
                        {
                            "answer": "4020",
                            "key": "GRWtfV",
                            "title": "Both revenue and capital",
                            "type": "text",
                        },
                        {
                            "answer": "This-is-a-type-text-answer",
                            "key": "SRWtfV",
                            "title": "Testing hyphen in field type text",
                            "type": "text",
                        },
                        {
                            "answer": "4020",
                            "key": "zvPzXN",
                            "title": "Revenue for 1 April 2024 to 31 March 2025",
                            "type": "text",
                        },
                        {
                            "answer": "1230",
                            "title": "Capital for 1 April 2023 to 31 March 2024",
                            "type": "text",
                        },
                        {
                            "answer": "1230",
                            "key": "pppiYl",
                            "title": "Capital for 1 April 2024 to 31 March 2025",
                            "type": "text",
                        },
                        {
                            "key": "NENGMj",
                            "title": "Which membership organisations are you a member of?",
                            "type": "list",
                            "answer": "homeless-link",
                        },
                        {
                            "key": "bCXNtj",
                            "title": "When did you start providing day provision?",
                            "type": "monthYear",
                            "answer": "2023-03",
                        },
                    ],
                    "question": "How much funding are you applying for?",
                },
                {
                    "fields": [
                        {
                            "answer": [
                                {
                                    "dpDFgB": "Test Funding Required NS Form",
                                    "iZdZrr": 40,
                                    "leIxEX": "1 April 2023 to 31 March 2024",
                                    "TrTaZQ": "None",
                                }
                            ],
                            "key": "mCbbyN",
                            "title": "Revenue costs",
                            "type": "multiInput",
                        }
                    ],
                    "question": "Revenue funding",
                },
                {
                    "fields": [
                        {
                            "answer": [
                                {
                                    "JtBjFp": 50,
                                    "cpFthG": "Test Funding Required NS Form",
                                    "mmwzGc": "1 April 2024 to 31 March 2025",
                                    "pMffVz": "Test Funding Required NS Form",
                                }
                            ],
                            "key": "XsAoTv",
                            "title": "Capital costs",
                            "type": "multiInput",
                        },
                        {
                            "key": "SxkwhF",
                            "type": "list",
                            "title": "Does your organisation have any alternative names?",
                            "answer": True,
                        },
                    ],
                    "question": "Capital funding",
                },
            ],
        }
    ],
    "form_names": [
        "funding-required-ns",
    ],
    "questions_answers": {
        "funding-required-ns": {
            "What funding are you applying for?": "Both revenue and capital",
            "What funding are you applying?": "Capital",
            "Partner organisation details": "version1 \n      . La la land, Mars, XA15 1AX\n      . https://www.wikipedia.org/\n      . Not provided\n      . Not provided",  # noqa
            "Partner's Job": ". version2: La la la land, Moon, XA15 1AL",
            "Both revenue and capital": "4020",
            "Testing hyphen in field type text": "This-is-a-type-text-answer",
            "Revenue for 1 April 2024 to 31 March 2025": "4020",
            "Capital for 1 April 2023 to 31 March 2024": "1230",
            "Capital for 1 April 2024 to 31 March 2025": "1230",
            "Which membership organisations are you a member of?": "Homeless link",
            "When did you start providing day provision?": "March 2023",
            "Revenue costs": "Test Funding Required NS Form \n      . 40\n      . 1 April 2023 to 31 March 2024\n      . Not provided",  # noqa
            "Capital costs": "Test Funding Required NS Form \n      . 50\n      . 1 April 2024 to 31 March 2025",
            "Does your organisation have any alternative names?": "Yes",
        }
    },
    "incorrect_form_data": [
        {
            "questions": [
                {
                    "category": "ZbxIUV",
                    "question": "Lead contact details",
                    "fields": [
                        {
                            "key": "fUMWcd",
                            "type": "text",
                            "answer": "test name",
                        },
                    ],
                }
            ],
            "name": "applicant-information-ns",
        },
    ],
    "exception_message": ("Could not map the data for form: applicant-information-ns"),
}

test_data_sort_questions_answers_welsh = {
    "forms": [
        {
            "name": "test_form",
            "questions": [
                {
                    "fields": [
                        {
                            "key": "SxkwhF",
                            "type": "list",
                            "title": "Os ydych yn llwyddiannus, a wnewch chi ddefnyddio eich cyllid o fewn y 12 mis nesaf?",
                            "answer": True,
                        },
                        {
                            "key": "SxkwhF",
                            "type": "list",
                            "title": "A ydych wedi sicrhau unrhyw gyllid cydweddu eto?",
                            "answer": False,
                        },
                    ],
                    "question": "Capital funding",
                },
            ],
        }
    ],
    "questions_answers": {
        "test_form": {
            "Os ydych yn llwyddiannus, a wnewch chi ddefnyddio eich cyllid o fewn y 12 mis nesaf?": "Oes",
            "A ydych wedi sicrhau unrhyw gyllid cydweddu eto?": "Nac Oes",
        }
    },
}


iso_and_nested_data = {
    "input_data": [
        {"HpLJyL__month": 3, "HpLJyL__year": 2022},
        {
            "addressLine1": "test",
            "addressLine2": "",
            "county": "",
            "postcode": "te3 2nr",
            "town": "test",
        },
        "wwww.example.com",
        None,
        None,
    ],
    "expected_response": "March 2022, test te3 2nr test, wwww.example.com",
}



"""
Data models for authentication
"""
from dataclasses import dataclass
from os import getenv
from typing import List
from typing import Mapping

from sentry_sdk import set_user

from .utils import get_highest_role_map


@dataclass
class User:
    full_name: str
    email: str
    roles: List[str]
    highest_role_map: Mapping[str, str]

    @classmethod
    def set_with_token(cls, token_payload):
        full_name = token_payload.get("fullName")
        email = token_payload.get("email")
        roles = token_payload.get("roles")
        # Set user in Sentry
        if getenv("SENTRY_DSN"):
            set_user(
                {
                    "email": token_payload.get("email"),
                    "id": token_payload.get("accountId"),
                    "username": token_payload.get("fullName"),
                }
            )
        return cls(
            full_name=full_name,
            email=email,
            roles=roles,
            highest_role_map=get_highest_role_map(roles),
        )



from typing import Mapping

import jwt as jwt
from flask import current_app

from .config import config_var_rs256_public_key


def _validate_token(token, key, algorithms):
    return jwt.decode(token, key, algorithms=algorithms)


def validate_token_rs256(token):
    key = current_app.config.get(config_var_rs256_public_key)
    return _validate_token(token, key, algorithms=["RS256"])


_ROLE_HIERARCHY = [
    "LEAD_ASSESSOR",
    "ASSESSOR",
    "COMMENTER",
]


def get_highest_role_map(roles: list[str]) -> Mapping[str, str]:
    roles_with_fund_prefix = tuple(f"_{rh}" for rh in _ROLE_HIERARCHY)
    filtered_roles = [r for r in roles if r.endswith(roles_with_fund_prefix)]

    fund_short_name_to_roles_list = {}
    for role in filtered_roles:
        fund_short_name, sub_role = role.split("_", 1)
        if fund_short_name in fund_short_name_to_roles_list:
            fund_short_name_to_roles_list[fund_short_name].append(sub_role)
        else:
            fund_short_name_to_roles_list[fund_short_name] = [sub_role]

    fund_short_name_to_highest_role = {}
    for fund_short_name, roles_list in fund_short_name_to_roles_list.items():
        highest_role, *_ = sorted(roles_list, key=lambda x: _ROLE_HIERARCHY.index(x))
        fund_short_name_to_highest_role[fund_short_name] = highest_role

    return fund_short_name_to_highest_role
